import cv2
import pystreaming.video.interface as intf
import types

# handler (function or generator): Any object that yield or return frames in BGR format.
def display(handler, BGR=True):
    if isinstance(handler, types.GeneratorType):
        for frame in handler:
            if not BGR:
                frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
            cv2.imshow("pystreaming-display", frame)
            if cv2.waitKey(1) & 0xFF == 27:  # esc pressed
                break
    else:
        while True:
            if not BGR:
                frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
            frame = handler()
            cv2.imshow("pystreaming-display", frame)
            if cv2.waitKey(1) & 0xFF == 27:  # esc pressed
                break



import zmq, zmq.asyncio, asyncio
import multiprocessing as mp


async def aioreq(context, source, track, drain):
    socket = context.socket(zmq.REQ)
    socket.connect(source)
    print(f"Requesting from {source}")
    while True:
        # rate limit to 30x a second => 90x requests a sec by default
        await asyncio.sleep(1 / 30)
        await socket.send(bytes(track, "utf-8"))
        buf = await socket.recv()
        idx = await socket.recv_pyobj()
        if idx == -1:
            continue  # throw away if no frame available
        print(3)
        await drain.send(buf, copy=False, flags=zmq.SNDMORE)
        await drain.send_pyobj(idx)


async def stop(shutdown):
    while True:
        await asyncio.sleep(0.1)  # Check 10x a second
        if shutdown.is_set():
            raise StopAsyncIteration()


def aiomain(source, track, outfd, procs, shutdown):
    context = zmq.asyncio.Context()
    drain = context.socket(zmq.PUSH)
    drain.bind(outfd)
    args = [aioreq(context, source, track, drain) for _ in range(procs)]
    args.append(stop(shutdown))
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(asyncio.gather(*args))
    except Exception:
        loop.stop()
        context.destroy(linger=0)
        loop.close()
    print("Exiting Requester")


class Requester:
    outfd = "ipc:///tmp/decin"

    def __init__(self, source, track="none", procs=3):
        self.source, self.procs, self.track = source, procs, track
        self.shutdown = mp.Event()
        self.psargs = (self.source, self.track, self.outfd, self.procs, self.shutdown)
        self.ps = None

    def start(self):
        if self.ps is not None:
            raise RuntimeError("Tried to start a runnning AIOREQ obj")
        self.ps = mp.Process(target=aiomain, args=self.psargs)
        self.ps.daemon = True
        print("RIGHT")
        self.ps.start()
        print("LEFT")

    def stop(self):
        if self.ps is None:
            raise RuntimeError("Tried to stop a stopped AIOREQ obj")
        self.shutdown.set()
        self.ps.join()
        self.ps = None
        self.shutdown.clear()



from turbojpeg import TurboJPEG, TJSAMP_420, TJFLAG_FASTDCT, TJFLAG_FASTUPSAMPLE
import numpy as np
import zmq, time
import multiprocessing as mp
from PIL import Image
from functools import partial
from itertools import count
from pystreaming.video.testimages.imageloader import loadimage
import pystreaming.video.interface as intf


def enc_ps(shutdown, infd, outfd, rcvhwm, outhwm):
    print("Starting Encoder")
    context = zmq.Context()
    socket = context.socket(zmq.PULL)
    socket.setsockopt(zmq.RCVHWM, rcvhwm)
    socket.connect(infd)
    out = context.socket(zmq.PUSH)
    out.setsockopt(zmq.SNDHWM, outhwm)
    out.connect(outfd)
    poller = zmq.Poller()
    poller.register(socket, zmq.POLLIN)
    encoder = partial(
        TurboJPEG().encode, quality=70, jpeg_subsample=TJSAMP_420, flags=TJFLAG_FASTDCT
    )
    while not shutdown.is_set():
        time.sleep(0.01)  # 100 cycles/sec
        if poller.poll(0):
            try:
                arr, idx = intf.recv_ndarray_idx(socket, flags=zmq.NOBLOCK)
                intf.send_buf_idx(out, encoder(arr), idx, flags=zmq.NOBLOCK)
            except zmq.error.Again:
                # ignore send misses to distributor.
                # Should not happen if there is a distributor.
                # Could implement a signal that there is a distributor. Not strictly necessary tho.
                pass
    print("Stopping Encoder")


class Encoder:
    infd = "ipc:///tmp/encin"
    outfd = "ipc:///tmp/encout"
    tellhwm = 30
    rcvhwm = outhwm = 10

    def __init__(self, context, procs=2):
        self.context, self.procs = context, procs
        self.shutdown = mp.Event()
        self.psargs = (self.shutdown, self.infd, self.outfd, self.rcvhwm, self.outhwm)
        self.workers = []
        self.idx = 0

        self.sender = self.context.socket(zmq.PUSH)
        self.sender.setsockopt(zmq.SNDHWM, self.tellhwm)
        self.sender.bind(self.infd)

    def send(self, frame):
        if self.workers == []:
            raise RuntimeError("Tried to send frame to stopped Encoder")
        try:
            intf.send_ndarray_idx(self.sender, frame, self.idx, flags=zmq.NOBLOCK)
            self.idx += 1
        except zmq.error.Again:
            raise RuntimeError("Worker processes are not processing frames fast enough")
        # change behavior to:
        # IF working, but slow, print a warning
        # If not responding, only then raise runtime error

    def start(self):
        if self.workers != []:
            raise RuntimeError("Tried to start running Encoder")
        for _ in range(self.procs):
            self.workers.append(mp.Process(target=enc_ps, args=self.psargs))

        for ps in self.workers:
            ps.daemon = True
            ps.start()

        time.sleep(2)  # Allow workers to spin up

    def stop(self):
        """Stop a pool of encoders.

        Raises:
            RuntimeError: Raised when the Encoder is already stopped
        """
        if self.workers == []:
            raise RuntimeError("Tried to stop stopped Encoder")

        self.shutdown.set()
        for ps in self.workers:
            ps.join()
        self.workers = []
        self.shutdown.clear()

    def _testcard(self, cardenum, animated=False):
        testimage = loadimage(cardenum)
        if self.workers == []:
            self.start()
        if animated:
            storage = []
            for ang in range(360):
                storage.append(np.asarray(testimage.rotate(ang)))
        else:
            storage = np.asarray(testimage)
        for i in count():
            print(i)
            self.send(storage[i % 360] if animated else storage)
            time.sleep(1 / 30)



from turbojpeg import TurboJPEG, TJFLAG_FASTDCT, TJFLAG_FASTUPSAMPLE
import zmq, time
from functools import partial
import pystreaming.video.interface as intf
import multiprocessing as mp


def dec_ps(shutdown, infd, outfd, rcvhwm, outhwm):
    print("Starting Decoder")
    context = zmq.Context()
    socket = context.socket(zmq.PULL)
    socket.setsockopt(zmq.RCVHWM, rcvhwm)
    socket.connect(infd)
    out = context.socket(zmq.PUSH)
    out.setsockopt(zmq.SNDHWM, outhwm)
    out.connect(outfd)
    poller = zmq.Poller()
    poller.register(socket, zmq.POLLIN)
    decoder = partial(TurboJPEG().decode, flags=(TJFLAG_FASTDCT + TJFLAG_FASTUPSAMPLE))

    while not shutdown.is_set():
        time.sleep(0.01)  # 100 cycles a sec
        if poller.poll(0):
            try:
                buf, idx = intf.recv_buf_idx(socket, flags=zmq.NOBLOCK)
                intf.send_ndarray_idx(out, decoder(buf), idx, flags=zmq.NOBLOCK)
            except zmq.error.Again:
                # ignore send misses to out.
                pass

    print("Stopping Decoder")


class Decoder:
    infd = "ipc:///tmp/decin"
    outfd = "ipc:///tmp/decout"
    hearhwm = 30
    rcvhwm = outhwm = 10

    def __init__(self, context, procs=2):
        self.context, self.procs = context, procs
        self.shutdown = mp.Event()
        self.psargs = (self.shutdown, self.infd, self.outfd, self.rcvhwm, self.outhwm)
        self.workers = []

        self.receiver = self.context.socket(zmq.PULL)
        self.receiver.bind(self.outfd)

    def recv(self):
        if self.workers == []:
            raise RuntimeError("Tried to receive frame from stopped Decoder")
        # Add timeout function?
        return intf.recv_ndarray_idx(self.receiver)

    def start(self):
        if self.workers != []:
            raise RuntimeError("Tried to start running Decoder")
        for _ in range(self.procs):
            self.workers.append(mp.Process(target=dec_ps, args=self.psargs))
        for ps in self.workers:
            ps.daemon = True
            ps.start()

        time.sleep(2)

    def stop(self):
        if self.workers == []:
            raise RuntimeError("Tried to stop stopped Decoder")

        self.shutdown.set()
        for ps in self.workers:
            ps.join()
        self.workers = []
        self.shutdown.clear()



from turbojpeg import TurboJPEG, TJSAMP_420, TJFLAG_FASTDCT, TJFLAG_FASTUPSAMPLE
from turbojpeg import TurboJPEG, TJFLAG_FASTDCT, TJFLAG_FASTUPSAMPLE
import numpy as np
from functools import partial
import cv2, blosc
from pystreaming.video.testimages.imageloader import loadimage

frame = np.asarray(loadimage(5))
decoder = partial(TurboJPEG().decode, flags=(TJFLAG_FASTDCT + TJFLAG_FASTUPSAMPLE))
# range(0, 101, 10):
for q in [50]:
    input(f"Testing {q=}")
    encoder = partial(
        TurboJPEG().encode, quality=q, jpeg_subsample=TJSAMP_420, flags=TJFLAG_FASTDCT
    )
    buf = encoder(frame)
    print(len(buf))
    recover = decoder(buf)
    cv2.imshow("window", recover)
    cv2.waitKey(0)


# x = np.asarray(loadimage(5))
# y = np.random.randint(0, 50, x.shape, dtype=np.uint8)
# # y = y - (y % 10)
# # y = np.zeros(x.shape, dtype=np.uint8)

# cv2.imshow("xxx", x)
# cv2.waitKey(0)
# cv2.imshow("xxx", y)
# cv2.waitKey(0)

# a = encoder(x)
# b = encoder(y)
# x = decoder(a)
# y = decoder(b)
# print(f"{len(a)=}")
# print(f"{len(b)=}")

# cv2.imshow("xxx", x)
# cv2.waitKey(0)
# cv2.imshow("xxx", y)
# cv2.waitKey(0)

# cap = cv2.VideoCapture(0)

# frm = [None, None]

# import time

# for i in range(90):
#     frame = cap.read()[1]
#     if i == 20 or i == 75:
#         frm[i // 40] = frame
#     print(i)

# s = time.time()
# diff = (frm[1]//2 + 128) - (frm[0]//2) 
# eend = time.time()
# print(f"TIME{eend - s}")

# cv2.imshow("xxx", frm[0])
# cv2.waitKey(0)
# cv2.imshow("xxx", frm[1])
# cv2.waitKey(0)
# cv2.imshow("xxx", diff)
# cv2.waitKey(0)


# s = time.time()
# e1 = encoder(frm[0])
# e2 = encoder(frm[1])
# end = time.time()
# print(f"DOUBLEENC:{end - s}")
# edelta = encoder(diff)

# e1rec = decoder(e1)
# e2rec = decoder(e2)
# edeltarec = decoder(edelta)
# import pdb; pdb.set_trace()
# e2tmpc = (edeltarec.astype(np.int32))*2 - 256 + e1rec
# e2tmpc = e2tmpc.astype(np.uint8)

# cv2.imshow("xxx", e1rec)
# cv2.waitKey(0)
# cv2.imshow("xxx", e2rec)
# cv2.waitKey(0)
# cv2.imshow("xxx", e2tmpc)
# cv2.waitKey(0)



# print(len(e1))
# print(len(blosc.compress(e1, typesize=8)))
# print(len(e2))
# print(len(edelta))


# encoder1 = partial(
#     TurboJPEG().encode, quality=50, jpeg_subsample=TJSAMP_420, flags=TJFLAG_FASTDCT
# )
# decoder = partial(TurboJPEG().decode, flags=(TJFLAG_FASTDCT + TJFLAG_FASTUPSAMPLE))


import cv2 
import time
import pystreaming as stream
import zmq
import numpy as np

def gstreamer_pipeline(
    capture_width=1920,
    capture_height=1080,
    display_width=1920,
    display_height=1080,
    framerate=30,
    flip_method=0,
):
    return (
        "nvarguscamerasrc ! "
        "video/x-raw(memory:NVMM), "
        "width=(int)%d, height=(int)%d, "
        "format=(string)NV12, framerate=(fraction)%d/1 ! "
        "nvvidconv flip-method=%d ! "
        "video/x-raw, width=(int)%d, height=(int)%d, format=(string)BGRx ! "
        "videoconvert ! "
        "video/x-raw, format=(string)BGR ! appsink"
        % (
            capture_width,
            capture_height,
            framerate,
            flip_method,
            display_width,
            display_height,
        )
    )

n = 10000

def encdistmain(cam):
    context = zmq.Context()
    bank = stream.Encoder(context, procs=2)
    dist = stream.Distributor("tcp://*:5553")
    bank.start()
    dist.start()
    bank._testcard(stream.TEST_M)
    # start = time.time()
    # for i in range(n):
    #     print(i)
    #     bank.send(cam.read()[1])
    #     time.sleep(1/30)
    # print(f"Did {n} frames in {time.time() - start}, fps = {n / (time.time() - start)}")
    # print("STOPPING")
    # bank.stop_workers()
    # dist.stop()

def recvmain():
    context = zmq.Context()
    req = stream.Requester("tcp://172.16.0.49:5553")
    dec = stream.Decoder(context, procs=2)
    req.start()
    dec.start()
    while True:
        frame, idx = dec.recv()
        cv2.imshow("test", frame)
        if cv2.waitKey(1) & 0xFF == ord('q'): 
            break
        print(f"Finally got idx={idx}")

class fakecamera:
    def __init__(self, size):
        self.frame = cv2.imread(f"./pystreaming/video/testimages/{size}.png")
    def read(self):
        time.sleep(1/35)
        return None, self.frame

def getstandcam(gst=False):
    if gst:
        cap = cv2.VideoCapture(gstreamer_pipeline(
            capture_width=640,
            capture_height=480,
            display_width=640,
            display_height=480,
            flip_method=2,
        ), cv2.CAP_GSTREAMER)
    else:
        cap = cv2.VideoCapture(0)
        cap.set(cv2.CAP_PROP_FOURCC, cv2.VideoWriter_fourcc(*"MJPEG"))
        cap.set(cv2.CAP_PROP_FRAME_WIDTH, 1920)
        cap.set(cv2.CAP_PROP_FRAME_HEIGHT, 1080)
        cap.set(cv2.CAP_PROP_FPS, 30)
    return cap


def yielder(animated=True):
    testimage = stream.loadimage(5)
    storage = []
    for ang in range(360):
        storage.append(np.asarray(testimage.rotate(ang)))
    i = 0
    if animated:
        while True:
            time.sleep(1/30)
            yield storage[i % 360]
            i += 1
    else:
        while True:
            time.sleep(1/30)
            yield storage[0]

if __name__ == "__main__":
    x = fakecamera("640x480_c")
    
    stream.display(yielder(), BGR=False)
    
    # encdistmain(fakecamera("1920x1080"))
    # recvmain()